
// CAD: Ctrl Alt Delete
// Copyright (C) 2007 S.Fuchita

#include <avr/io.h>
#include <avr/interrupt.h>


#define REPEAT_RATE 13
#define REPEAT_DELAY 200

#include "queue16.h"
#include "ps2if.h"
#include "typedef.h"

typedef struct Queue16 Queue;
#define qinit(Q) q16init(Q)
#define qpeek(Q) q16peek(Q)
#define qget(Q) q16get(Q)
#define qput(Q,C) q16put(Q,C)
#define qunget(Q,C) q16unget(Q,C)



enum {
	KEYFLAG_MAKE = 0x80,
};


Queue event_queue_l;
Queue event_queue_h;
Queue pc_input_queue;
Queue pc_output_queue;

unsigned short pc_input_bits;
unsigned short pc_output_bits;

unsigned short repeat_timeout;



// spin timer for 8MHZ clock

void waitloop(unsigned int n);


#define wait40us()		waitloop(40)
#define wait15us()		waitloop(14)
#define wait24us()		waitloop(24)


int countbits(unsigned short c)
{
	int i;
	for (i = 0; c; c >>= 1) {
		if (c & 1) i++;
	}
	return i;
}

// initialize

inline void initialize_pc_output()
{
	qinit(&pc_output_queue);
	pc_output_bits = 0;
}

inline void initialize_pc_input()
{
	qinit(&pc_input_queue);
	pc_input_bits = 0;
}

bool pc_send(unsigned char c)
{
	unsigned char i;
	unsigned short bits;

	bits = c & 0xff;
	if (!(countbits(bits) & 1)) bits |= 0x100;
	bits = (bits | 0x200) << 1;

	for (i = 0; i < 11; i++) {
		if (!pc_get_clock()) break;
		if (bits & 1)	pc_set_data_1();
		else			pc_set_data_0();
		bits >>= 1;
		wait15us();
		if (!pc_get_clock()) break;
		pc_set_clock_0();
		wait40us();
		pc_set_clock_1();
		wait15us();
	}
	pc_set_data_1();

	return i == 11;
}		

int pc_recv()
{
	int i;
	unsigned short bits;

	if (!pc_get_clock()) {
		return -1;
	}
	if (pc_get_data()) {
		return -1;
	}

	bits = 0;
	for (i = 0; i < 100; i++) {
		pc_set_clock_0();
		wait40us();
		pc_set_clock_1();
		wait15us();
		wait15us();
		if (i < 9) {
			bits >>= 1;
			if (pc_get_data()) bits |= 0x100;
		} else {
			if (pc_get_data()) {	// stop bit ?
				break;
			}
		}
	}
	pc_set_clock_0();
	pc_set_data_0();
	wait40us();
	pc_set_data_1();
	pc_set_clock_1();

	if (i != 9) {
		return -1;							// framing error
	}

	if (!(countbits(bits & 0x1ff) & 1)) {
		return -1;							// parity error
	}

	return bits & 0xff;
}





inline void pc_put(unsigned char c)
{
	qput(&pc_output_queue, c & 0xff);
}

inline int pc_get()
{
	int c = qget(&pc_input_queue);
	return c;
}

inline unsigned short get_event()
{
	unsigned short l, h;
	cli();
	l = qget(&event_queue_l);
	h = qget(&event_queue_h);
	sei();
	if ((l | h) & 0xff00) return 0;
	return (h << 8) | l;
}

inline void put_event(unsigned short c)
{
	cli();
	qput(&event_queue_l, c);
	qput(&event_queue_h, c >> 8);
	sei();
}

inline void unget_event(unsigned short c)
{
	cli();
	qunget(&event_queue_l, c);
	qunget(&event_queue_h, c >> 8);
	sei();
}




unsigned short count10 = 0;
bool timerevent = false;



unsigned char keydata0;
unsigned char keydata1;
unsigned char keydata2;


// 10ms intervals
SIGNAL(SIG_OUTPUT_COMPARE0A)
{
	TCNT0 = 0; 
	timerevent = true;

	unsigned char d = PINB;
	keydata1 = keydata2 | d;
	keydata2 = d;

}




enum {
	PC_INIT			= 0x0000,
	PC_NORMAL		= 0x0100,
	PC_ED			= 0x0200,
	PC_F0			= 0x0300,
	PC_F3			= 0x0400,
	PC_FB			= 0x0500,
	PC_FC			= 0x0600,
	PC_FD			= 0x0700,
	PC_FF			= 0x0800,
};

enum {
	KB_INIT			= 0x0000,
	KB_NORMAL		= 0x0100,
	KB_WAIT			= 0x0200,
	KB_ED			= 0x0300,
	KB_F3			= 0x0400,
};

enum {
	EVENT_NULL					= 0,
	EVENT_INIT					= 0x0100,
	EVENT_KEYMAKE				= 0x0200,
	EVENT_KEYBREAK				= 0x0300,
};



unsigned short pc_state;
unsigned short pc_init_count;
bool scan_enable;

static void _send_pc(unsigned char c, void *cookie)
{
	pc_put(c);
}

void handler()
{
	int c;
	long event;

	if (timerevent) {
		timerevent = false;
		if (pc_state == PC_INIT) {
			if (pc_init_count > 0) {
				pc_init_count--;
			} else {
				initialize_pc_output();
				pc_put(0xaa);
				scan_enable = true;
				pc_state = PC_NORMAL;
			}
		}
	}

	event = get_event();
	if (event > 0) {
		switch (event & 0xff00) {
		case EVENT_INIT:
			scan_enable = false;
			pc_init_count = 40;		// wait 400ms
			pc_state = PC_INIT;
			break;
		case EVENT_KEYMAKE:
			c = event & 0xff;
			ps2encode(c, KEYFLAG_MAKE, _send_pc, 0);
			break;
		case EVENT_KEYBREAK:
			c = event & 0xff;
			ps2encode(c, 0, _send_pc, 0);
			break;
		}
	}

	c = pc_get();
	if (c >= 0) {
		switch (pc_state & 0xff00) {
		case PC_ED:
			pc_put(0xfa);
			pc_state = PC_NORMAL;
			break;
		case PC_F0:
			if (!c) {
				pc_put(0x02);	// scan code set 2
			} else {
				pc_put(0xfa);
			}
			pc_state = PC_NORMAL;
			break;
		case PC_F3:
			pc_put(0xfa);
			pc_state = PC_NORMAL;
			break;
		case PC_INIT:
			if (c == 0xec) {
				break;
			}
			pc_state = PC_NORMAL;	// cancel initialize command
			scan_enable = true;
			// fallthru
		default:
			switch (c) {
			case 0xed:
				pc_put(0xfa);
				pc_state = PC_ED;
				break;
			case 0xee:
				pc_put(0xee);
				break;
			case 0xf0:
				pc_put(0xfa);
				pc_state = PC_F0;
				break;
			case 0xf1:
				pc_put(0xfa);
				break;
			case 0xf2:
				pc_put(0xfa);
				pc_put(0xab);
				pc_put(0x83);
				scan_enable = true;
				break;
			case 0xf3:			// set typematic rate/delay
				pc_put(0xfa);
				pc_state = PC_F3;
				break;
			case 0xf4:			// enable
				pc_put(0xfa);
				scan_enable = true;
				break;
			case 0xf5:			// default disable
				pc_put(0xfa);
				scan_enable = false;
				break;
			case 0xf6:
			case 0xf7:
			case 0xf8:
			case 0xf9:
			case 0xfa:
			case 0xfb:
			case 0xfc:
			case 0xfd:
				pc_put(0xfa);
				break;
			case 0xfe:
				break;
			case 0xff:
				pc_put(0xfa);
				put_event(EVENT_INIT);
				scan_enable = false;
				break;
			default:
				pc_put(0xfe);
				break;
			}
		}
	}

	{
		unsigned char d = keydata1;
		unsigned char x = keydata0 ^ d;
		keydata0 = d;
		if (x & 1) {	// ctrl
			if (d & 1) {
				put_event(EVENT_KEYBREAK | 58);
			} else {
				put_event(EVENT_KEYMAKE | 58);
			}
		}
		if (x & 2) {	// alt
			if (d & 2) {
				put_event(EVENT_KEYBREAK | 60);
			} else {
				put_event(EVENT_KEYMAKE | 60);
			}
		}
		if (x & 4) {	// delete
			if (d & 4) {
				put_event(EVENT_KEYBREAK | 76);
			} else {
				put_event(EVENT_KEYMAKE | 76);
			}
		}
	}
}


int main()
{
	int c;

	PORTD = 0;
	DDRD = 0x2a;

	UBRRH = 0;
	UBRRL = 12;
	UCSRB |= 0x08;		// transmitter enable

	PORTB = 0;
	DDRB = 0xc0;

	qinit(&event_queue_l);
	qinit(&event_queue_h);
	initialize_pc_input();
	initialize_pc_output();

	pc_state = PC_INIT;
	pc_init_count = 0;
	scan_enable = true;

	keydata0 = 0;
	keydata1 = 0;
	keydata2 = 0;

	// timer enable

	TCCR0B = 5;					// prescaler set to clk/1024
	OCR0A = 79;					// about 10ms intervals
	TIMSK |= 1 << OCIE0A;		// enable output compare interrupt
	sei();						// enable global interrupt

	// event loop

	while (1) {
		handler();

		// receive from host
		c = pc_recv();
		if (c >= 0) {
			qput(&pc_input_queue, c);
		}

		// transmit to host
		c = qget(&pc_output_queue);
		if (c >= 0) {
			if (!pc_send(c)) {
				qunget(&pc_output_queue, c);	// failed
			}
		}
	}
}

